/*
 * Copyright 2023, Exotic Silicon, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. This software is licensed exclusively under this specific license text.  The license text may not be changed, and the software
 *    including modified versions may not be re-licensed under any other license text.
 * 2. Redistributions of source code must retain the above copyright notice, this list of conditions, and the following disclaimer.
 * 3. Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the following
 *    disclaimer in the documentation and/or other materials provided with the distribution.
 * 4. All advertising materials mentioning features or use of this software must display the following acknowledgement: This product
 *    includes software developed by Exotic Silicon.
 * 5. The name of Exotic Silicon must not be used to endorse or promote products derived from this software without specific prior
 *    written permission.
 * 6. Redistributions of modified versions of the source code must be clearly identified as having been modified from the original.
 * 7. Redistributions in binary form that have been created from modified versions of the source code must clearly state in the
 *    documentation and/or other materials provided with the distribution that the source code has been modified from the original.
 *
 * THIS SOFTWARE IS PROVIDED 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EXOTIC SILICON BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Text screendump utility, version 1.0.
 *
 * For more information about this utility, and the kernel patches it requires, please visit:
 * https://research.exoticsilicon.com/articles/console_screendumps
 *
 * Define ENHANCED_CONSOLE if running on OpenBSD 7.3-release or later, or -current sources after 20230118.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <dev/wscons/wsconsio.h>
#include <fcntl.h>
#include <dev/wscons/wsdisplayvar.h>

#define BLUE(x) (48*((x-16)%6))
#define GREEN(x) (48*(((x-16)/6)%6))
#define RED(x) (48*(((x-16)/36)%6))
#define HEX(i) ((i & 0xf) <10 ? (i & 0xf)+'0' : (i & 0xf)-10+'a')
#define GREY(i) (1+((i-232)*11))

void colour_html(int cindex, unsigned char * return_col)
{
if (cindex < 16) {
	*(return_col+3)=0;
	*return_col=(cindex & 1 ? (cindex & 8 ? 'f' : '8') : '0');
	*(return_col+1)=(cindex & 2 ? (cindex & 8 ? 'f' : '8') : '0');
	*(return_col+2)=(cindex & 4 ? (cindex & 8 ? 'f' : '8') : '0');
	if (cindex==7) {
		*(return_col)='d';
		*(return_col+1)='d';
		*(return_col+2)='d';
		}
	if (cindex==8) {
		*(return_col)='8';
		*(return_col+1)='8';
		*(return_col+2)='8';
		}
	return;
	}
*(return_col+6)=0;
if (cindex < 232) {
	*(return_col)=HEX(RED(cindex) >> 4);
	*(return_col+1)=HEX(RED(cindex));
	*(return_col+2)=HEX(GREEN(cindex) >> 4);
	*(return_col+3)=HEX(GREEN(cindex));
	*(return_col+4)=HEX(BLUE(cindex) >> 4);
	*(return_col+5)=HEX(BLUE(cindex));
	return ;
	}
*(return_col)=HEX(GREY(cindex) >> 4);
*(return_col+1)=HEX(GREY(cindex));
*(return_col+2)=HEX(GREY(cindex) >> 4);
*(return_col+3)=HEX(GREY(cindex));
*(return_col+4)=HEX(GREY(cindex) >> 4);
*(return_col+5)=HEX(GREY(cindex));
return ;
}

int main()
{
int fd;
int res;
int i, j;
int pos_in;
int spanopen;
unsigned int cattr;
struct wsdisplay_screendump_text screendump;
unsigned char * buffer;
unsigned char * pos_out;
unsigned char * return_col;

screendump.textdata=malloc(32768);
screendump.attrdata=malloc(65536);

res=ioctl (STDIN_FILENO, WSDISPLAYIO_SCREENDUMP_TEXT, &screendump);

if (res == -1) {
	printf ("Ioctl WSDISPLAYIO_SCREENDUMP_TEXT not supported for this device\n");
	return (1);
	}

printf ("ioctl returned %d\nrows %d, cols %d\n", res, screendump.rows, screendump.cols);

/* Write raw text output to a file */
fd=open ("screendump.raw", O_WRONLY | O_CREAT | O_TRUNC, 0600);
write (fd, screendump.textdata, screendump.rows * screendump.cols);
close (fd);

/* Write line-delimited output to a file */
buffer=malloc(screendump.rows * (screendump.cols + 1));
j=0;
for (i=0 ; i < screendump.rows * screendump.cols ; i++) {
	*(buffer+j++)=*(screendump.textdata+i);
	if ((i+1) % screendump.cols == 0) {
		while (j>0 && *(buffer + j - 1)==32) { j--; }
		*(buffer+j++)=10;
		}
	}
fd=open ("screendump.asc", O_WRONLY | O_CREAT | O_TRUNC, 0600);
write (fd, buffer, j);
close (fd);
free (buffer);

/* Write HTML output to a file */
buffer=malloc(1024*1024);
pos_out=buffer;

spanopen=0;

#define LINEAR_POS (i*screendump.cols+j)

/* Write HTML preamble */
pos_out+=sprintf(pos_out, "<html><body><div style='font-family:monospace;white-space:pre;'>");

return_col=malloc(1024);

/* Process the output line-by-line */
for (i=0; i<screendump.rows; i++) {
	pos_out+=sprintf(pos_out, "<div>");
	for (j=0; j<screendump.cols; j++) {
		if (j==0 || screendump.attrdata[LINEAR_POS]!=cattr) {
			cattr=screendump.attrdata[LINEAR_POS];
			if (j>0) {
				pos_out+=sprintf(pos_out, "</span>");
				}
			pos_out+=sprintf(pos_out, "%s", "<span style='");
			colour_html((cattr >> 16) & 0xff, return_col);
			pos_out+=sprintf(pos_out, "background:#%s;", return_col);
			colour_html((cattr >> 24) & 0xff, return_col);
			pos_out+=sprintf(pos_out, "color:#%s;", return_col);
#ifdef ENHANCED_CONSOLE
			pos_out+=sprintf(pos_out, "font-weight:%d;", (cattr & WSATTR_HILIT ? 800 : 400));
			pos_out+=sprintf(pos_out, "text-decoration:%s;", (cattr & WSATTR_UNDERLINE ? "underline" : "none"));
#else
			pos_out+=sprintf(pos_out, "text-decoration:%s;", (cattr & 1 ? "underline" : "none"));
#endif
			pos_out+=sprintf(pos_out, "%s", "'>");
			}
		switch (*(screendump.textdata+LINEAR_POS)) {
			case '>':
				pos_out+=sprintf(pos_out, "%s", "&gt;");
				break;
			case '<':
				pos_out+=sprintf(pos_out, "%s", "&lt;");
				break;
			case '&':
				pos_out+=sprintf(pos_out, "%s", "&amp;");
				break;
			default:
				*(pos_out++)=*(screendump.textdata+LINEAR_POS);
			}
		}
	pos_out+=sprintf(pos_out, "</span></div>");
	}

/* Write HTML trailer */
pos_out+=sprintf(pos_out, "</div></body></html>");

fd=open ("screendump.html", O_WRONLY | O_CREAT | O_TRUNC, 0600);
write (fd, buffer, pos_out-buffer);
close (fd);
free (buffer);
free (return_col);
free (screendump.textdata);
free (screendump.attrdata);
return (0);
}
